/**
@file PCF8583.c
@brief Funktionsbibliothek zum ansteuern einer PCF8583 Real Time Clock per TWI
@author Marc Riedlinger
@date 23.05.2009
@version 1.0
*/

#include "PCF8583.h"
#include "i2cmaster.h"
#include <inttypes.h>

/**Adresse der RTC*/
#define RTC_ADDR	0xA0

/**Jahr in dem die Zeitrechnung beginnt (aktuelles Jahr = YEAR_ZERO + 4 x DELTA_YEAR + YY), muss Schaltjahr sein*/
#define YEAR_ZERO	1900

//Speicheraddressen der Register des PCF8583
//Control/Status Register
#define CSR_ADDR 	0x00
//Hundertstel Sekunden
#define HOS_ADDR 	0x01
//Sekunden
#define SEC_ADDR 	0x02
//Minuten
#define MIN_ADDR	0x03
//Stunden
#define HOUR_ADDR 	0x04
//Jahr + Tag
#define YYDD_ADDR	0x05
//Wochentag + Monat
#define WDMM_ADDR	0x06
//Anzahl vergangener 4-Jahres-Perioden
#define DELTA_YEAR_ADDR	0x10
//Alarm Minuten
#define MIN_ALERT_ADDR 0x0B
//Alert Control Register
#define ACR_ADDR 0x08


//Bitmasken
//Stop Counting Flag
#define SCF_MASK 	0x80
//Hold last count flag
#define HLCF_MASK 	0x40
//Function Mode Select
#define FM_MASK		0x30
//Mask Flag
#define MF_MASK		0x08
//Alert Enable
#define AE_MASK		0x04
//Alert Flag
#define AF_MASK		0x02
//Timer Flag
#define TF_MASK		0x01

//1/10s
#define S1_10_MASK	0xF0
//1/100s
#define	S1_100_MASK	0x0F
//10s
#define S10_MASK	0xF0
//1s
#define S1_MASK		0x0F
//10 min
#define MIN10_MASK	0xF0
//1 min
#define MIN1_MASK	0x0F
//10 h
#define H10_MASK	0x30
//1 h
#define H1_MASK		0x0F
//AM/PM Flag
#define AMPM_MASK	0x40
//Hour Format
#define HF_MASK		0x80
//1 Tag
#define DD1_MASK	0x0F
//10 Tag
#define DD10_MASK	0x30
//Jahr
#define JJ_MASK		0xC0
//1 Monat
#define MM1_MASK	0x0F
//10 Monat
#define MM10_MASK	0x10
//Wochentag
#define WD_MASK		0xE0
//Clock Alert Function
#define CAF_MASK 	0x30
//Timer Alert Function
#define TAF_MASK	0x40
//Alert Interrupt Enable
#define AIE_MASK 	0x80




//Bitpositionen
//Stop Counting Flag
#define SCF_POS 	7
//Hold last count flag
#define HLCF_POS 	6
//Function Mode Select
#define FM_POS		4
//Mask Flag
#define MF_POS		3
//Alert Enable
#define AE_POS		2
//Alert Flag
#define AF_POS		1
//Timer Flag
#define TF_POS		0

//1/10s
#define S1_10_POS	4
//1/100s
#define	S1_100_POS	0
//10s
#define S10_POS		4
//1s
#define S1_POS		0
//10 min
#define MIN10_POS	4
//1 min
#define MIN1_POS	0
//10 h
#define H10_POS		4
//1 h
#define H1_POS		0
//AM/PM Flag
#define AMPM_POS	6
//Hour Format
#define HF_POS		7
//1 Tag
#define DD1_POS		0
//10 Tag
#define DD10_POS	4
//Jahr
#define JJ_POS		6
//1 Monat
#define MM1_POS		0
//10 Monat
#define MM10_POS	4
//Wochentag
#define WD_POS		5

//Alert-Control-Register
//Clock Alert Function
#define CAF_POS 	4
//Timer Alert Function
#define TAF_POS		6
//Alert Interrupt Enable
#define AIE_POS 	7



//Betriebsmodi der RTC
//Clock Mode mit 32768 Hz Takt
#define MODE_CM32 0b00
//Clock Mode mit 50 Hz Takt
#define MODE_CM50 0b01


//Alarmtypen
//kein Alarm
#define ALERT_OFF 0b00
//tglicher Alarm
#define ALERT_DAILY 0b01
//wochentagspezifischer Alarm
#define ALERT_WEEKDAY 0b01
//datumsspezifischer Alarm
#define ALERT_DATE 0b11



/**
@brief initialisiert die RTC
@param mode	Modus in der die RTC betrieben wird
@retval 0 RTC nicht erreichbar
@retval 1 RTC korrekt initialisiert
*/
uint8_t rtc_init(uint8_t mode){
	uint8_t status;
	//I2C Initialisiern
	i2c_init();
	//Mode setzen
	status = i2c_start(RTC_ADDR + I2C_WRITE);
	status = i2c_write(CSR_ADDR);
	status = i2c_write((mode<<FM_POS));
	i2c_stop();
	return (!status);
}

/**
@details setzt den Sekundenwert in der RTC
@param	sec	zu setzender Sekundenwert
@retval 0 RTC nicht erreichbar
@retval	1 Wert gesetzt
*/
uint8_t rtc_setSeconds(uint8_t sec){
	sec = rtc_intToBCD(sec);
	rtc_writeRegister(HOS_ADDR, 0);
	return rtc_writeRegister(SEC_ADDR, sec);

}

/**
@details setzt den Minutenwert in der RTC
@param	min	zu setzender Minutenwert
@retval 0 RTC nicht erreichbar
@retval	1 Wert gesetzt
*/
uint8_t rtc_setMinutes(uint8_t min){
	min = rtc_intToBCD(min);
	return rtc_writeRegister(MIN_ADDR, min);
}

/**
@details setzt den Stundenwert in der RTC im 24h Format
@param	hour	zu setzender Stundenwert
@retval 0 RTC nicht erreichbar
@retval	1 Wert gesetzt
*/
uint8_t rtc_setHours24(uint8_t hour){
	hour = rtc_intToBCD(hour);
	return rtc_writeRegister(HOUR_ADDR, hour);
}

/**
@details fragt den aktuellen Sekundenwert von der RTC ab
@retval 0..59	aktuelle Sekunde
*/
uint8_t rtc_getSeconds(){
	uint8_t sec;
	sec = rtc_readRegister(SEC_ADDR);
	sec = rtc_bcdToInt(sec);
	return sec;
}


/**
@details fragt den aktuellen Minutenwert von der RTC ab
@retval 0..59	aktuelle Minute
*/
uint8_t rtc_getMinutes(){
	uint8_t min;
	min = rtc_readRegister(MIN_ADDR);
	min = rtc_bcdToInt(min);
	return min;
}

/**
@details fragt den aktuellen Stundenwert von der RTC im 24h format ab
@retval 0..23	aktuelle Stunde
*/
uint8_t rtc_getHours24(){
	uint8_t hour;
	hour = rtc_readRegister(HOUR_ADDR);
	hour &=	(H1_MASK | H10_MASK);
	hour = rtc_bcdToInt(hour);
	return hour;
}

/**
@details setzt den Stundenzhler auf einen bestimmten Wert
@param days	aktueller Tag
@retval 0	bertragung fehlerhaft
@retval
*/
uint8_t rtc_setDays(uint8_t days){
	//setzt den Tag ohne die Jahreszahl zu
	return rtc_registerAND(rtc_intToBCD(days), YYDD_ADDR, (DD1_MASK | DD10_MASK));
}



/**
@details fragt den aktuellen Tageswert von der RTC ab
@retval 0..31	aktueller Tag
*/
uint8_t rtc_getDays(){
	uint8_t days;
	//Register mit Tag lesen
	days = rtc_readRegister(YYDD_ADDR);
	//Jahr abschneiden
	days &=	((DD1_MASK) | (DD10_MASK));
	days = rtc_bcdToInt(days);
	return days;
}



/**
@details fragt den aktuellen Monat von der RTC ab
@retval 0..12	aktueller Monat
*/
uint8_t rtc_getMonth(){
	uint8_t month;
	//Register mit Tag lesen
	month = rtc_readRegister(WDMM_ADDR);
	//Wochentag abschneiden
	month &= ((MM1_MASK) | (MM10_MASK));
	month = rtc_bcdToInt(month);
	return month;
}



/**
@details fragt das aktuelle Jahr von der RTC ab
@retval 1900..2924	aktuelles Jahr
*/
uint16_t rtc_getYear(){
	uint16_t year;
	//Register mit Jahr auslesen
	year = (uint16_t) rtc_readRegister(YYDD_ADDR);
	year = (year>>JJ_POS);
	//Offset auslesen
	year += (((uint16_t) rtc_readRegister(DELTA_YEAR_ADDR))*4);
	//Startjahr hinzuaddieren
	year +=	YEAR_ZERO;
	return year;
}



/**
@details setzt das aktuelle Jahr
@param year	aktuelles Jahr
@retval 0	bertragung fehlerhaft
@retval	1	bertragung erfolgreich
*/
uint8_t rtc_setYear(uint16_t year){
	year -= YEAR_ZERO;
	rtc_writeRegister(DELTA_YEAR_ADDR, (year/4));
	return rtc_registerAND(((year%4)<<JJ_POS), YYDD_ADDR, JJ_MASK);
}


/**
@details setzt den aktuellen Monat
@param days	aktueller Monat
@retval 0	bertragung fehlerhaft
@retval	1	bertragung erfolgreich
*/
uint8_t rtc_setMonth(uint8_t month){
	//setzt den Monat ohne den Wochentag zu ndern
	return rtc_registerAND(rtc_intToBCD(month), WDMM_ADDR, (MM1_MASK | MM10_MASK));
}


/**
@details fragt den aktuellen Wochentag von der RTC ab
@retval 0..6	aktueller Wochentag	(0 = Montag)
*/
uint8_t rtc_getWeekday(){
	uint8_t weekday;
	//Register mit Wochentag lesen
	weekday = rtc_readRegister(WDMM_ADDR);
	//Monat abschneiden
	weekday = (weekday>>WD_POS);
	return weekday;
}


/**
@details setzt den aktuellen Wochentag
@param days	aktueller Wochentag	(0 = Montag)
@retval 0	bertragung fehlerhaft
@retval	1	bertragung erfolgreich
*/
uint8_t rtc_setWeekday(uint8_t weekday){
	//setzt den Wochentag ohne den Monat zu ndern
	return rtc_registerAND((weekday<<WD_POS), WDMM_ADDR, WD_MASK);
}



/**
@details wandelt eine Integerzahl in seine BCD-Darstellung um
@param	zu wandelnde Integerzahl
@retval	BCD-Darstellung: ZZZZEEEE (Z=10er, E=1er)
*/
uint8_t rtc_intToBCD(uint8_t intVal){
	uint8_t bcd;
	bcd = ((intVal/10)<<4);
	bcd |= (intVal % 10);
	return bcd;
}

/**
@details wandelt eine BCD-Darstellung in einen Integer um
@param	zu wandelnde BCD-Zahl
@retval	Integer des BCD Wertes
*/
uint8_t rtc_bcdToInt(uint8_t bcd){
	uint8_t intVal;
	intVal = (bcd>>4)*10;
	intVal += (bcd & 0x0F);
	return intVal;
}


/**
@details setzt nur die durch eine Bitmaske selektierten Bits eines Registers der RTC auf einen neuen Wert
@param	val		neuer Inhalt des Registers
@param	regAddr	Speicheradresse des Registers
@param	bitmask	selektiert die Bits die aktualisiert werden sollen (logisch 1 = selektiert)
@retval 0 Fehler bei bertragung
@retval 1 bertragung erfolgreich
*/
uint8_t rtc_registerAND(uint8_t val, uint8_t regAddr, uint8_t bitmask){
	uint8_t status, oldReg;
	status = i2c_start(RTC_ADDR + I2C_WRITE);
	//Leseadresse setzen
	status = i2c_write(regAddr);
	//Slave Transmitter
	status = i2c_rep_start(RTC_ADDR + I2C_READ);
	//aktuellen Registerinhalt lesen
	oldReg = i2c_readNak();
	//selektierte Bits 0 setzten
	oldReg &= ~(bitmask);
	//selektierte Bits mit neuen Werten fllen
	oldReg |= val;
	//Master Transmitter
	status = i2c_rep_start(RTC_ADDR + I2C_WRITE);
	//Schreibadresse setzen
	status = i2c_write(regAddr);
	//aktualisierten Registerinhalt senden
	status = i2c_write(oldReg);
	i2c_stop();
	return (!status);
}

/**
@details Setzt eine Alarmzeit zu der tglich ein Interrupt ausgegeben wird. Alarm muss erst ueber rtc_enableAlert() aktiviert werden!
@param	hour 	Stunde des Alarms im 24h Format
@param	minute	Minute des Alarm
@retval 0 Fehler bei bertragung
@retval 1 bertragung erfolgreich
*/
uint8_t rtc_setDailyAlert(uint8_t hour, uint8_t minute){
	uint8_t regContent[2];
	regContent[0] = rtc_bcdToInt(minute);
	regContent[1] = rtc_bcdToInt(hour);
	//Alarmzeit auf RTC schreiben
	rtc_writeRegisterSequence(MIN_ALERT_ADDR, 2, regContent);
	//Modus tglicher Alarm + Interrupt bei Alarm
	return rtc_registerAND(((ALERT_DAILY<<CAF_POS) | (1<<AIE_POS)), ACR_ADDR, (CAF_MASK | AIE_MASK));
}


/**
@details Aktiviert einen gesetzten Alarm
@retval 0 Fehler bei bertragung
@retval 1 bertragung erfolgreich
*/
uint8_t rtc_enableAlert(){
	//Alert Enable im CS-Register + Alert Flag lschen
	return rtc_registerAND((1<<AE_POS), CSR_ADDR, (AE_MASK | AF_MASK));
}

/**
@details	Liest ein Register der RTC aus
@param	Speicheraddresse des auszulesenden Registers
@retval	Inhalt des Registers
*/
uint8_t rtc_readRegister(uint8_t regAddr){
	uint8_t content;
	i2c_start(RTC_ADDR + I2C_WRITE);
	//Schreibadresse setzen
	i2c_write(regAddr);
	//Slave Transmitter
	i2c_rep_start(RTC_ADDR + I2C_READ);
	//Sekundenwert lesen
	content = i2c_readNak();
	i2c_stop();
	return content;
}

/**
@details Schreibt in ein Register der RTC
@param	regAddr	Speicheradresse des Registers
@param	data	Neuer Wert fr das Register
@retval 0 Fehler bei bertragung
@retval 1 bertragung erfolgreich
*/
uint8_t rtc_writeRegister(uint8_t regAddr, uint8_t data){
	uint8_t status;
	status = i2c_start(RTC_ADDR + I2C_WRITE);
	//Schreibadresse setzen
	status = i2c_write(regAddr);
	//Wert setzen
	status = i2c_write(data);
	i2c_stop();
	return (!status);
}



/**
@details Schreibt in mehrere aufeinanderfolgende Register der RTC
@param	startAddr	Speicheradresse des ersten Registers
@param	*data		Zeiger auf das erste Element des Arrays mit den neuen Werten
@param	numReg		Anzahl der zu beschreibenden Register
@retval 0 Fehler bei bertragung
@retval 1 bertragung erfolgreich
*/
uint8_t rtc_writeRegisterSequence(uint8_t startAddr, uint8_t numReg, uint8_t *data){
	uint8_t status, i;
	status = i2c_start(RTC_ADDR + I2C_WRITE);
	//Schreibadresse setzen
	status = i2c_write(startAddr);
	for (i = 0; i<numReg; i++){
		//Wert setzen
		status = i2c_write(*data);
		data++;
	}
	i2c_stop();
	return (!status);
}

